@@ -279,6 +279,48 @@ class GitClientTest(GitClientTestSetups):
         client = GitClient(self.remote_path)
         self.assertEquals('', client.get_status())
 
+
+class GitClientUpdateTest(GitClientTestSetups):
+
+    def test_update_fetch_all_tags(self):
+        url = self.remote_path
+        client = GitClient(self.local_path)
+        self.assertTrue(client.checkout(url, "master"))
+        self.assertEqual(client.get_branch(), "master")
+        self.assertTrue(client.update())
+        p = subprocess.Popen("git tag", shell=True, cwd=self.local_path, stdout=subprocess.PIPE)
+        output = p.communicate()[0].decode('utf-8')
+        self.assertEqual('last_tag\ntest_tag\n', output)
+        subprocess.check_call("git checkout test_tag", shell=True, cwd=self.remote_path)
+        subprocess.check_call("git branch alt_branch", shell=True, cwd=self.remote_path)
+        subprocess.check_call("touch alt_file.txt", shell=True, cwd=self.remote_path)
+        subprocess.check_call("git add *", shell=True, cwd=self.remote_path)
+        subprocess.check_call("git commit -m altfile", shell=True, cwd=self.remote_path)
+        # switch to untracked
+        subprocess.check_call("git checkout test_tag", shell=True, cwd=self.remote_path)
+        subprocess.check_call("touch new_file.txt", shell=True, cwd=self.remote_path)
+        subprocess.check_call("git add *", shell=True, cwd=self.remote_path)
+        subprocess.check_call("git commit -m newfile", shell=True, cwd=self.remote_path)
+        subprocess.check_call("git tag new_tag", shell=True, cwd=self.remote_path)
+        self.assertTrue(client.update())
+        # test whether client gets the tag
+        p = subprocess.Popen("git tag", shell=True, cwd=self.local_path, stdout=subprocess.PIPE)
+        output = p.communicate()[0].decode('utf-8')
+        self.assertEqual('''\
+last_tag
+new_tag
+test_tag
+''', output)
+        p = subprocess.Popen("git branch -a", shell=True, cwd=self.local_path, stdout=subprocess.PIPE)
+        output = p.communicate()[0].decode('utf-8')
+        self.assertEqual('''\
+* master
+  remotes/origin/HEAD -> origin/master
+  remotes/origin/alt_branch
+  remotes/origin/master
+  remotes/origin/test_branch
+''', output)
+
 class GitClientLogTest(GitClientTestSetups):
 
     def setUp(self):
@@ -329,6 +371,14 @@ class GitClientDanglingCommitsTest(GitClientTestSetups):
         po = subprocess.Popen("git log -n 1 --pretty=format:\"%H\"", shell=True, cwd=self.local_path, stdout=subprocess.PIPE)
         self.untracked_version = po.stdout.read().decode('UTF-8').rstrip('"').lstrip('"')
 
+        # diverged branch
+        subprocess.check_call("git checkout test_tag -b diverged_branch", shell=True, cwd=self.local_path)
+        subprocess.check_call("touch diverged.txt", shell=True, cwd=self.local_path)
+        subprocess.check_call("git add *", shell=True, cwd=self.local_path)
+        subprocess.check_call("git commit -m diverged_branch", shell=True, cwd=self.local_path)
+        po = subprocess.Popen("git log -n 1 --pretty=format:\"%H\"", shell=True, cwd=self.local_path, stdout=subprocess.PIPE)
+        self.diverged_branch_version = po.stdout.read().decode('UTF-8').rstrip('"').lstrip('"')
+
         # Go detached to create some dangling commits
         subprocess.check_call("git checkout test_tag", shell=True, cwd=self.local_path)
         # create a commit only referenced by tag
@@ -336,6 +386,9 @@ class GitClientDanglingCommitsTest(GitClientTestSetups):
         subprocess.check_call("git add *", shell=True, cwd=self.local_path)
         subprocess.check_call("git commit -m no_branch", shell=True, cwd=self.local_path)
         subprocess.check_call("git tag no_br_tag", shell=True, cwd=self.local_path)
+        po = subprocess.Popen("git log -n 1 --pretty=format:\"%H\"", shell=True, cwd=self.local_path, stdout=subprocess.PIPE)
+        self.no_br_tag_version = po.stdout.read().decode('UTF-8').rstrip('"').lstrip('"')
+
         # create a dangling commit
         subprocess.check_call("touch dangling.txt", shell=True, cwd=self.local_path)
         subprocess.check_call("git add *", shell=True, cwd=self.local_path)
@@ -344,9 +397,24 @@ class GitClientDanglingCommitsTest(GitClientTestSetups):
         po = subprocess.Popen("git log -n 1 --pretty=format:\"%H\"", shell=True, cwd=self.local_path, stdout=subprocess.PIPE)
         self.dangling_version = po.stdout.read().decode('UTF-8').rstrip('"').lstrip('"')
 
+        # create a dangling tip on top of dangling commit (to catch related bugs)
+        subprocess.check_call("touch dangling-tip.txt", shell=True, cwd=self.local_path)
+        subprocess.check_call("git add *", shell=True, cwd=self.local_path)
+        subprocess.check_call("git commit -m dangling_tip", shell=True, cwd=self.local_path)
+
+        # create and delete branch to cause reflog entry
+        subprocess.check_call("git branch oldbranch", shell=True, cwd=self.local_path)
+        subprocess.check_call("git branch -D oldbranch", shell=True, cwd=self.local_path)
+
         # go back to master to make head point somewhere else
         subprocess.check_call("git checkout master", shell=True, cwd=self.local_path)
 
+    def test_is_commit_in_orphaned_subtree(self):
+        client = GitClient(self.local_path)
+        self.assertTrue(client.is_commit_in_orphaned_subtree(self.dangling_version))
+        self.assertFalse(client.is_commit_in_orphaned_subtree(self.no_br_tag_version))
+        self.assertFalse(client.is_commit_in_orphaned_subtree(self.diverged_branch_version))
+
     def test_protect_dangling(self):
         client = GitClient(self.local_path)
         # url = self.remote_path
@@ -473,6 +541,26 @@ class GitClientDanglingCommitsTest(GitClientTestSetups):
         except VcsError:
             pass
 
+class GitClientOverflowTest(GitClientTestSetups):
+    '''Test reproducing an overflow of arguments to git log'''
+
+    def setUp(self):
+        client = GitClient(self.local_path)
+        client.checkout(self.remote_path)
+        subprocess.check_call("git checkout test_tag", shell=True, cwd=self.local_path)
+        subprocess.check_call("echo 0 >> count.txt", shell=True, cwd=self.local_path)
+        subprocess.check_call("git add count.txt", shell=True, cwd=self.local_path)
+        subprocess.check_call("git commit -m modified-0", shell=True, cwd=self.local_path)
+        # produce many tags to make git log command fail if all are added
+        for count in range(4000):
+            subprocess.check_call("git tag modified-%s" % count, shell=True, cwd=self.local_path)
+        po = subprocess.Popen("git log -n 1 --pretty=format:\"%H\"", shell=True, cwd=self.local_path, stdout=subprocess.PIPE)
+        self.last_version = po.stdout.read().decode('UTF-8').rstrip('"').lstrip('"')
+
+    def test_orphaned_overflow(self):
+        client = GitClient(self.local_path)
+        # this failed when passing all ref ids to git log
+        self.assertFalse(client.is_commit_in_orphaned_subtree(self.last_version))
 
 class GitDiffStatClientTest(GitClientTestSetups):
 
